""" 
# ===========================================================================
# Interface between the ION's implementation of the CCSDS LTP and Python. 
# Internally, all classes call the C Extension _ltp.
#
# Author: Marc Sanchez Net
# Date:   05/30/2019
# Copyright (c) 2019, California Institute of Technology ("Caltech").  
# U.S. Government sponsorship acknowledged.
# ===========================================================================
"""

# General imports
from unittest.mock import Mock
from pathlib import Path
from threading import Event, Thread
from warnings import warn
import contextlib

# Module imports
from . import utils
from .mgmt import sm_task_yield

# Import C Extension
try:
    from . import _ltp
except ImportError:
    from unittest.mock import Mock
    _ltp = Mock()

# Define all methods/vars exposed at pyion
__all__ = [
    'ltp_init',
    'ltp_dequeue_outbound_segment',
    'ltp_handle_inbound_segment',
    'AccessPoint',
]

# ============================================================================
# === Segment queueing functionality
# ============================================================================

def ltp_init(estMaxExportSessions):
    _ltp.ltp_init(estMaxExportSessions)

@contextlib.contextmanager
def ltp_dequeue_outbound_segment(vspan):
    try:
        yield _ltp.ltp_dequeue_outbound_segment(vspan)
    finally:
        # Close every LTP dequeue segment handling block with a task yield
        sm_task_yield()

def ltp_handle_inbound_segment(data):
    _ltp.ltp_handle_inbound_segment(data)

# ============================================================================
# === AccessPoint class
# ============================================================================

class AccessPoint():
    """ Class to represent an LTP access point. """
    def __init__(self, proxy, client_id, sap_addr):
        # Store variables
        self.proxy     = proxy
        self.client_id = client_id
        self._sap_addr = sap_addr
        self.node_dir  = proxy.node_dir
        self._result   = None

    def __del__(self):
        # If you have already been closed, return
        if not self.is_open:
            return

        # Close the Entity and free C memory
        self.proxy.ltp_close(self.client_id)

    @property
    def is_open(self):
        """ Returns True if the access point is opened """
        return (self.proxy is not None and self._sap_addr is not None)

    def _cleanup(self):
        """ Clean access point after closing. Do not call directly,
            use ``proxy.ltp_close``
        """
        # Clear variables
        self.proxy     = None
        self.client_id = None
        self._sap_addr = None

    @utils._chk_is_open
    @utils.in_ion_folder
    def ltp_send(self, dest_engine_nbr, data):
        """ Trigger LTP to send data 

            :param: Destination engine number
            :param: Data as str, bytes or bytearray
        """
        self._result = _ltp.ltp_send(self._sap_addr, dest_engine_nbr, data)

    @utils._chk_is_open
    def ltp_receive(self):
        """ Trigger LTP to receive data """
        # Receive on another thread. Otherwise you cannot handle a SIGINT
        th = Thread(target=self._ltp_receive, daemon=True)
        th.start()
        th.join()

        # If exception, raise it
        if isinstance(self._result, Exception):
            raise self._result

        return self._result
        
    @utils.in_ion_folder
    def _ltp_receive(self):
        # Get payload from next bundle. If exception, return it too
        try:
            self._result = _ltp.ltp_receive(self._sap_addr)
        except Exception as e:
            self._result = e
            
    @utils._chk_is_open
    @utils.in_ion_folder
    def ltp_interrupt(self):
        """ Trigger LTP to be interrupted """
        _ltp.ltp_interrupt(self._sap_addr)

    def __enter__(self):
        """ Allows an endpoint to be used as context manager """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Allows an endpoint to be used as context manager """
        pass

    def __str__(self):
        return '<AccessPoint: {} ({})>'.format(self.client_id, 'Open' if self.is_open else 'Closed')

    def __repr__(self):
        return '<AccessPoint: {} ({})>'.format(self.client_id, self._sap_addr)



""" 
# ===========================================================================
# **pyion** provides a Python interface to JPL's Interplantary Overlay Network 
# (ION) via a Proxy class that allows create Endpoints through which you can
# send or receive bytes, bytearrays and memoryviews. They operate in a 
# manner similar to how TCP/UDP sockets operate.
#
# **pyion** also provides similar abstractions to interface with
# - LTP
# - CFDP
# - SDR and PSM
# - Management functions such as add/delete a contact
#
# **NONE OF THE PRIMITIVES PROVIDED IN PYION ARE THREAD-SAFE**
#
# **pyion** does not substitute traditional ION management functions such as ``ionadmin``
# or ``bpadmin`` since only a small subset functionality is implemented. Therefore, to
# use pyion, it is required that
# (1) ION be installed in the host computer
# (2) One or multiple ION nodes are already instantiated and running.
#
# **pyion's** documentation is available at https://pyion.readthedocs.io/en/latest/
#
# Author: Marc Sanchez Net
# Date:   04/17/2019
# Copyright (c) 2019, California Institute of Technology ("Caltech").  
# U.S. Government sponsorship acknowledged.
# ===========================================================================
"""

# General imports
import os
import warnings    

# This pulls up import so that from the user you can do
# ``import pyion`` instead of having to do one of the following:
#   - ``from pyion import pyion``
#   - ``import pyion.pyion as pyion``
from .proxies import *
from .utils import *
from .constants import *

# Handle case where _mgmt was not compiled during setup
try:
    from .mgmt import *
except ImportError:
    warnings.warn("Pyion's management module is not installed.")

# Must be set if multiple ION nodes are run on the same host.
ION_NODE_LIST_DIR = None


""" 
# ===========================================================================
# Interface between the ION's implementation of the Bundle Protocol and Python. 
# Internally, all classes call the C Extension _bp.
#
# Author: Marc Sanchez Net
# Date:   04/17/2019
# Copyright (c) 2019, California Institute of Technology ("Caltech").  
# U.S. Government sponsorship acknowledged.
# ===========================================================================
"""

# General imports
from unittest.mock import Mock
import os
from pathlib import Path
from threading import Thread
from warnings import warn

# Module imports
from . import utils
from .constants import BpEcsEnumeration, BpCustodyEnum

# Import C Extension
from . import _bp

# Define all methods/vars exposed at pyion
__all__ = ['Endpoint']

# ============================================================================
# === Endpoint object
# ============================================================================

class Endpoint():
	""" Class to represent and Endpoint. Do not instantiate it manually, use
		the Proxy's ``open``, ``close``, ``interrupt`` methods instead.

		:ivar proxy: Proxy that created this endpoint.
		:ivar eid:   Endpoint ID.
		:ivar node_dir: Directory for this node. Inherited from the proxy that
						created this endpoint object.
		:ivar _sap_addr: Memory address of the BpSapState object used by the 
						 C Extension. Do not modify or potential memory leak 						
	"""
	def __init__(self, proxy, eid, sap_addr, TTL, priority, report_eid,
				 custody, report_flags, ack_req, retx_timer, detained, 
				 chunk_size, timeout, mem_ctrl):
		""" Endpoint initializer  """
		# Store variables
		self.proxy        = proxy
		self.eid 	      = eid
		self._sap_addr    = sap_addr
		self.node_dir     = proxy.node_dir
		self.TTL          = TTL
		self.priority     = priority
		self.report_eid   = report_eid
		self.custody      = custody
		self.report_flags = report_flags
		self.ack_req 	  = ack_req
		self.retx_timer   = retx_timer
		self.chunk_size   = chunk_size
		self.detained     = detained
		self.timeout      = timeout
		self.mem_ctrl     = mem_ctrl

		# TODO: This properties are hard-coded because they are not 
		# yet supported
		self.sub_priority = 1	# Also known as ``ordinal``
		self.criticality  = ~int(BpEcsEnumeration.BP_MINIMUM_LATENCY)

		# Mark if the endpoint is blocked 
		self.tx_result = None
		self.rx_result = None

	def __del__(self):
		# Attempt to close prior to deleting
		self.close()

	@property
	def is_open(self):
		""" Returns True if this endpoint is opened """
		return (self.proxy is not None and self._sap_addr is not None)

	def _cleanup(self):
		""" Deletes eid and sap_addr information to indicate that this endpoint is closed

			.. Danger:: DO NOT call this function directly. Use the Proxy ``close``
						function instead, or the enpoint close
		"""
		self.eid  	   = None
		self._sap_addr = None
		self.proxy     = None

	def close(self):
		""" Close this endpoint if necessary. """
		if not self.is_open:
			return
		self.proxy.bp_close(self)

	@utils._chk_is_open
	@utils.in_ion_folder
	def bp_send(self, dest_eid, data, TTL=None, priority=None,
				report_eid=None, custody=None, report_flags=None,
				ack_req=None, retx_timer=None, chunk_size=None):
		""" Send data through the proxy

			:param dest_eid: Destination EID for this data
			:param data: Data to send as ``bytes``, ``bytearray`` or a ``memoryview``
			:param **kwargs: See ``Proxy.bp_open``
		"""
		# Get default values if necessary
		if TTL is None: TTL = self.TTL
		if priority is None: priority = self.priority
		if report_eid is None: report_eid = self.report_eid
		if custody is None: custody = self.custody
		if report_flags is None: report_flags = self.report_flags
		if ack_req is None: ack_req = self.ack_req
		if retx_timer is None: retx_timer = self.retx_timer
		if chunk_size is None: chunk_size = self.chunk_size

		# If this endpoint is not detained, you cannot use a retx_timer
		if retx_timer>0 and not self.detained:
			raise ConnectionError('This endpoint is not detained. You cannot set up custodial timers.')

		# In pyion v4.0.0 and beyond, BPv7 is assumed. Therefore, custody transfer is not allowed
		if custody != BpCustodyEnum.NO_CUSTODY_REQUESTED:
			raise ValueError('Custody transfer is not allowed in pyion-4.0.0+ since it is not in BPv7')

		# Create call arguments
		args = (dest_eid, data, TTL, priority, report_eid, custody, report_flags,
				ack_req, retx_timer, chunk_size)

		# Open another thread because otherwise you cannot handle a SIGINT if blocked
		# waiting for ZCO space
		th = Thread(target=self._bp_send, args=args, daemon=True)
		th.start()

		# Wait for a bundle to be sent
		th.join()

		# If sending resulted in an exception, raise it
		if isinstance(self.tx_result, BaseException):
			raise self.tx_result

	def _bp_send(self, dest_eid, data, TTL, priority, report_eid, custody, report_flags,
				ack_req, retx_timer, chunk_size):
		""" Send data through the proxy

			:param dest_eid: Destination EID for this data
			:param data: Data to send as ``bytes``, ``bytearray`` or a ``memoryview``
			:param **kwargs: See ``Proxy.bp_open``
		"""
		# If you need to send in full, do it
		if chunk_size is None:
			self._bp_send_bundle(dest_eid, data, TTL, priority, report_eid, custody,
				 				 report_flags, ack_req, retx_timer)
			return

		# If data is a string, then encode it to get a bytes object
		if isinstance(data, str): data = data.encode('utf-8')

		# Create a memoryview object
		memv = memoryview(data)

		# Send data in chuncks of chunk_size bytes
		# NOTE: If data is not a multiple of chunk_size, the memoryview
		#  		object returns the correct end of the buffer.
		for i in range(0, len(memv), chunk_size):
			self._bp_send_bundle(dest_eid, memv[i:(i+chunk_size)].tobytes(), TTL, 
								 priority, report_eid, custody, report_flags, 
								 ack_req, retx_timer)

	def _bp_send_bundle(self, dest_eid, data, TTL, priority, report_eid, custody,
				 		report_flags, ack_req, retx_timer):
		""" Transmit a bundle """
		try:
			_bp.bp_send(self._sap_addr, dest_eid, report_eid, TTL, priority,
						custody, report_flags, int(ack_req), retx_timer, data)
			self.tx_result = None
		except BaseException as e:
			self.tx_result = e

	def bp_send_file(self, dest_eid, file_path, **kwargs):
		""" Convenience function to send a file

			.. Tip:: To send file in chunk, use ``chunk_size`` argument

			:param str: Destination EID
			:param Path or str: Path to file being sent
			:param **kwargs: See ``Proxy.bp_open``
		"""
		# Initialize variables
		file_path = Path(file_path)

		# If file does not exist, raise error
		if not file_path.exists():
			raise IOError('{} is not valid'.format(file_path))

		# Send it
		self.bp_send(dest_eid, file_path.read_bytes(), **kwargs)

	def _bp_receive(self, chunk_size, return_headers):
		""" Receive one or multiple bundles.

			:param chunk_size: Number of bytes to receive at once.
			:param return_headers: If true, will return bundle headers.
		"""
		# If no chunk size defined, simply get data in the next bundle
		if chunk_size is None:
			self.rx_result = self._bp_receive_bundle(return_headers)
			return

		# Pre-allocate buffer of the correct size and create a memory view
		buf  = bytearray(chunk_size)
		memv = memoryview(buf)

		# Initialize variables
		extra_bytes = None
		bytes_read  = 0
		headers = {}

		# Read bundles as long as you have not reached the chunk size
		while self.is_open and bytes_read < chunk_size:
			# Get data from next bundle
			data = _bp.bp_receive(self._sap_addr, return_headers)
			
			# If data is of type exception, return
			if isinstance(data, BaseException):
				self.rx_result = data
				return

			# Create memoryview to this bundle's payload
			bmem = memoryview(data[0])
			sz   = len(bmem)

			if return_headers:
				# Extract headers from bundle
				headers = data[1]

			# Put as much data as possible in the pre-allocated buffer
			idx     = min(bytes_read+sz, chunk_size)
			to_read = idx-bytes_read
			memv[bytes_read:idx] = bmem[0:to_read]

			# If this bundle has data that does not fit in pre-allocated buffer
			# append it at then at the expense of a copy
			if to_read < sz:
				extra_bytes = bmem[to_read:]
			
			# Update the number of bytes read
			bytes_read += sz

		# If there are extra bytes add them
		
		self.rx_result = (memv.tobytes(), headers)
		if extra_bytes: self.rx_result[0] += extra_bytes.tobytes()
		
		if not return_headers:
			self.rx_result = self.rx_result[0]

	@utils.in_ion_folder
	def _bp_receive_bundle(self, return_headers):
		""" Receive one bundle """
		# Get payload from next bundle. If exception, return it too
		try:
			return _bp.bp_receive(self._sap_addr, return_headers)
		except BaseException as e:
			return e

	@utils._chk_is_open
	def bp_receive(self, chunk_size=None, timeout=None, return_headers=False):
		""" Receive data through the proxy. This is BLOCKING call. If an error
			occurred while receiving, an exception is raised.
		
			.. Tip:: The actual data reception is done in a separate thread
					 so that the user can interrupt it with a SIGINT. 
			.. Warning:: Use the chunk_size parameter with care. It is provided because each
						 ``bp_receive`` call has some overhead (e.g. a new thread is spawned),
						 so you might not want to do this if you expect very small bundles. 
						 However, if the tx sends an amount of data that is not a multiple of
						 chunk_size, then ``bp_receive`` will be blocked forever.

			:param chunk_size: If not specified, then read one bundle at a time, with
							   each bundle having a potentially different size.
							   If specified, it tells you the least number of bytes to read
							   at once (i.e., this function will return a bytes object
							   with length >= chunk_size)
			:param timeout: If not specified, then no timeout is used. If specified,
							reception of a bundle will be cancelled after timeout seconds.
			:param return_headers: If set to True, the return value of bp_receive will be a
									tuple where the first value is the payload and the second
									is a dictionary containing the bundle header information.
									This value defaults to False. 
		"""
		# Get default values if necessary
		if chunk_size is None: chunk_size = self.chunk_size
		if timeout is None: timeout = self.timeout

		# For now, do not let user use both chunk_size and timeout together
		# to avoid complexity
		if chunk_size and timeout:
			raise ValueError('bp_receive cannot have chunk_size and timeout at the same time.')

		# Open another thread because otherwise you cannot handle a SIGINT
		th = Thread(target=self._bp_receive, args=(chunk_size,return_headers,), daemon=True)
		th.start()

		# Wait for a bundle to be delivered	and set a timeout
		th.join(timeout)

		# Handle the case where you have timed-out
		if th.is_alive():
			self._bp_receive_timeout()
			return self.rx_result

		# If exception, raise it
		if isinstance(self.rx_result, BaseException):
			raise self.rx_result

		return self.rx_result

	def _bp_receive_timeout(self):
		# Interrupt the reception of bundles from ION. 
		self.proxy.bp_interrupt(self)

	def __enter__(self):
		""" Allows an endpoint to be used as context manager """
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		""" Allows an endpoint to be used as context manager """
		self.close()

	def __str__(self):
		return '<Endpoint: {} ({})>'.format(self.eid, 'Open' if self.is_open else 'Closed')

	def __repr__(self):
		return '<Endpoint: {} ({})>'.format(self.eid, self._sap_addr)




"""
EmionEngine — Python interface to ION-DTN via pyion C-bindings.
Handles memory attachment, bundle send/receive.
No dummies — requires a running ION node.
"""

import os
import time


class EmionEngine:
    """
    High-level Python wrapper over pyion C-bindings for a single ION node.
    Must be used after EmionNode.start() has booted the C daemons.
    """

    def __init__(self, node_id: int, base_dir: str = "/tmp/emion_nodes"):
        self.node_id = node_id
        self.node_dir = os.path.join(base_dir, str(node_id))
        self.proxy = None

    def attach(self, retries=5, delay=2):
        """Attach to the ION shared-memory region via pyion."""
        print(f"[EmION] Attaching engine to Node {self.node_id}...")

        # Deferred import to allow bootstrapping
        from emion import pyion

        # Ensure ION_NODE_LIST_DIR is unset — we use cwd instead
        os.environ.pop("ION_NODE_LIST_DIR", None)
        pyion.ION_NODE_LIST_DIR = None
        os.environ["ION_NODE_NUMBER"] = str(self.node_id)

        saved_dir = os.getcwd()
        os.chdir(self.node_dir)

        try:
            for attempt in range(1, retries + 1):
                try:
                    self.proxy = pyion.get_bp_proxy(self.node_id)
                    print(f"      ✅ Attached (attempt {attempt})")
                    return True
                except Exception as e:
                    print(f"      [!] Attempt {attempt}/{retries}: {e}")
                    time.sleep(delay)
        finally:
            os.chdir(saved_dir)

        raise RuntimeError(
            f"Cannot attach to ION Node {self.node_id}. "
            f"Is the node running? Check {self.node_dir}/node_boot.log"
        )

    def send(self, src_eid: str, dst_eid: str, payload: bytes):
        """Send a bundle via the real ION Bundle Protocol engine."""
        if not self.proxy:
            self.attach()
        with self.proxy.bp_open(src_eid) as ep:
            result = ep.bp_send(dst_eid, payload)
            print(f"      📡 Sent {len(payload)}B  {src_eid} → {dst_eid}")
            return result

    def receive(self, eid: str, timeout: int = 10):
        """Receive a bundle from the ION BP engine."""
        if not self.proxy:
            self.attach()
        with self.proxy.bp_open(eid) as ep:
            data = ep.bp_receive(timeout=timeout)
            if data:
                print(f"      📥 Received {len(data)}B on {eid}")
            return data

    def detach(self):
        """Detach from the ION proxy."""
        if self.proxy:
            self.proxy = None



"""
EmION CLI — all operations use real ION-DTN.

    emion dashboard     Launch the web dashboard
    emion setup         Install ION-DTN and pyion dependencies
    emion info          Show ION-DTN status
    emion test          Run the 2-node communication test
"""

import sys
import argparse


def main():
    parser = argparse.ArgumentParser(
        prog="emion",
        description="EmION — Authentic ION-DTN Framework"
    )
    sub = parser.add_subparsers(dest="cmd")

    d = sub.add_parser("dashboard", help="Launch the web dashboard")
    d.add_argument("--host", default="0.0.0.0")
    d.add_argument("--port", type=int, default=8420)

    sub.add_parser("setup", help="Install ION-DTN and build pyion bindings")
    sub.add_parser("build", help="Manually build/rebuild pyion C-bindings")
    sub.add_parser("info", help="Show ION-DTN status")
    sub.add_parser("test", help="Run 2-node comm test")
    
    args = parser.parse_args()

    if args.cmd == "dashboard":
        from emion.dashboard.server import run
        run(host=args.host, port=args.port)
    elif args.cmd == "setup":
        _setup()
    elif args.cmd == "build":
        _build()
    elif args.cmd == "info":
        _info()
    elif args.cmd == "test":
        _test()
    else:
        parser.print_help()


def _setup():
    import subprocess
    import os
    import shutil
    import tempfile

    print("\n  ⚛ EmION Standalone Universal Setup")
    print("  ====================================")
    
    # 1. Check for ION-DTN
    if not shutil.which("ionadmin"):
        print("  [1/1] ION-DTN not found. Starting Autonomous Build...")
        
        try:
            # Install system dependencies
            print("      (Installing system build tools...)")
            subprocess.run(["sudo", "apt-get", "update"], check=True)
            subprocess.run(["sudo", "apt-get", "install", "-y", 
                          "build-essential", "cmake", "automake", "libtool", "wget", "tar"], check=True)
            
            # Download and Build ION-DTN
            with tempfile.TemporaryDirectory() as tmpdir:
                ion_url = "https://github.com/NASA-AMMOS/ION-DTN/archive/refs/tags/v4.1.2.tar.gz"
                print(f"      (Downloading ION-DTN v4.1.2...)")
                subprocess.run(["wget", ion_url, "-O", os.path.join(tmpdir, "ion.tar.gz")], check=True)
                subprocess.run(["tar", "-xzf", os.path.join(tmpdir, "ion.tar.gz"), "-C", tmpdir], check=True)
                
                # Find the extracted directory
                extracted_dir = None
                for d in os.listdir(tmpdir):
                    if os.path.isdir(os.path.join(tmpdir, d)) and "ION-DTN" in d:
                        extracted_dir = os.path.join(tmpdir, d)
                        break
                
                if not extracted_dir:
                    raise Exception("Failed to find extracted ION-DTN directory.")
                
                print(f"      (Compiling & Installing ION-DTN... this may take a few minutes)")
                # Build steps
                subprocess.run(["./configure"], cwd=extracted_dir, check=True)
                subprocess.run(["make", "-j", str(os.cpu_count() or 2)], cwd=extracted_dir, check=True)
                subprocess.run(["sudo", "make", "install"], cwd=extracted_dir, check=True)
                subprocess.run(["sudo", "ldconfig"], check=True)
                
            print("  ✅ ION-DTN installed globally.")
        except Exception as e:
            print(f"  ❌ Failed to autonomously install ION-DTN: {e}")
            print("     Manual fix: Download ION-DTN v4.1.2, run ./configure && make && sudo make install")
            return
    else:
        print("  ✅ ION-DTN already installed.")

    print("\n  🎉 ION-DTN Setup Complete!")
    _build()


def _build():
    import subprocess
    import sys
    import os
    from pathlib import Path

    print("\n  ⚛ EmION C-Extension Build")
    print("  ===========================")
    
    # Try to find the root of the package to run build
    import emion
    pkg_dir = Path(emion.__file__).parent.parent
    
    # Check if we are in a source tree or installed package
    if os.path.exists(pkg_dir / "setup.py"):
        root_dir = pkg_dir
    else:
        # If installed, we might not have the source. 
        # But for EmION, the sdist includes everything.
        # However, 'pip install --force-reinstall' is the best way for installed users.
        print("  [!] Running from installed package. Attempting force-reinstall build...")
        try:
            subprocess.run([sys.executable, "-m", "pip", "install", "emion", "--force-reinstall", "--no-cache-dir"], check=True)
            print("  ✅ Bindings rebuilt successfully via pip.")
            return
        except Exception as e:
            print(f"  ❌ Failed to rebuild via pip: {e}")
            return

    print(f"  [1/1] Building extensions in {root_dir}...")
    try:
        # Build extensions inplace
        subprocess.run([sys.executable, "setup.py", "build_ext", "--inplace"], cwd=root_dir, check=True)
        print("  ✅ Bindings built successfully.")
    except Exception as e:
        print(f"  ❌ Build failed: {e}")
        print("     Ensure ION-DTN is installed and headers are in /usr/local/include")


def _info():
    import shutil
    import emion
    print(f"\n  ⚛ EmION v{emion.__version__}")
    
    ion_found = True
    for cmd in ["ionadmin", "bpadmin", "ionsecadmin", "bprecvfile", "bpsendfile"]:
        path = shutil.which(cmd)
        if not path: ion_found = False
        print(f"  {cmd}: {'✅ ' + path if path else '❌ not found'}")
    
    pyion_found = True
    try:
        from emion import pyion
        print(f"  pyion (internal): ✅ available")
    except ImportError as e:
        pyion_found = False
        print(f"  pyion (internal): ❌ not built or linked ({e})")
        print(f"  (Run: pip install emion --force-reinstall)")

    if not ion_found:
        print("\n  ⚠️  ION-DTN missing! Run 'emion setup' to fix.")
    
    try:
        import fastapi
        print(f"  fastapi: v{fastapi.__version__}")
    except ImportError:
        print(f"  fastapi: ❌  (pip install fastapi uvicorn websockets)")
    print()


def _test():
    from tests.test_emion import test_two_node_communication
    ok = test_two_node_communication()
    sys.exit(0 if ok else 1)


if __name__ == "__main__":
    main()



""" 
# ===========================================================================
# Module with utility functions and classes for pyion
#
# Author: Marc Sanchez Net
# Date:   04/17/2019
# Copyright (c) 2019, California Institute of Technology ("Caltech").  
# U.S. Government sponsorship acknowledged.
# ===========================================================================
"""

# Generic imports
from abc import ABC
from datetime import datetime, timedelta
from enum import Enum, unique
from functools import wraps
import os
from pathlib import Path
import time

from . import __init__ as pyion

# ============================================================================
# === Define variables/methods/classes that will be directly importable
# ============================================================================

__all__ = ['check_ion_env_vars']

# ============================================================================
# === Error messages
# ============================================================================

_env_var_missing = ('{} enviornment variable is not properly set',
                    'Current value is {}',
                    'You can either set it in bash using ``export ${}`` or',
                    'inside Python using ``pyion.{} = "..."``')

_path_missing = 'Path "{}" does not exist'

# ============================================================================
# === Abstract Proxy
# ============================================================================

class Proxy(ABC):
    """ Abstract proxy to ION. 

        Public Variables
        ----------------
        :ivar node_nbr:  Node number that this proxy is action upon.
        :ivar node_dir:  Directory for this node
        :ivar attached:  True if proxy is attached to ION.
    """
    def __init__(self, node_nbr):
        # Initialize variables
        self.node_nbr = node_nbr
        self.attached = False

        # Set the ION_NODE_LIST_DIR variable if necessary. This call resolves relative
        # paths, makes sure directory exits.
        set_ion_node_list_dir(pyion.ION_NODE_LIST_DIR)

        # Get ION_NODE_LIST_DIR environment variable
        nodes_path = os.environ.get('ION_NODE_LIST_DIR')

        # Set node number and node dir
        self.node_dir = Path(nodes_path)/str(node_nbr) if nodes_path else None

    def __str__(self):
        return '<{}: {} ({})>'.format(self.__class__.__name__, self.node_nbr,
                                      'Attached' if self.attached else 'Detached')

    def __repr__(self):
        return str(self)

# ============================================================================
# === Helper functions
# ============================================================================

def in_ion_folder(func):
    """ Decorator to change directory inside a node's folder to run ION
        commands. It can only be used from within the ``Proxy`` class
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # Get the node's directory
        node_dir = getattr(self, 'node_dir')

        # If no directory specified, just run. This is always the case
        # unless you run multiple ION nodes in a single machine.
        if node_dir is None: return func(self, *args, **kwargs)
    
        # Store current working directory
        cur_dir = os.getcwd()

        # Go to the node's directory
        os.chdir(str(node_dir.absolute()))

        # Execture function
        try:
            ret = func(self, *args, **kwargs)
        finally:
            # Go back to the previous working directory
            os.chdir(cur_dir)

        return ret
    return wrapper

def _chk_attached(func):
    """ Decorator that checks if Proxy is attached """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # You need to be attached
        if not self.attached:
            raise IOError('Not attached to ION, call ``attach`` first.')

        # Call function
        return func(self, *args, **kwargs)
    return wrapper

def _chk_is_open(func):
    """ Decorator that checks if PoP is opened """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # You need to be attached
        if not self.is_open:
            raise ConnectionAbortedError('{}'.format(self))

        # Call function
        return func(self, *args, **kwargs)
    return wrapper

def check_ion_env_vars(ION_NODE_LIST_DIR):
    """ Check the ION environment variables to ensure they are consistent (e.g., 
         the paths set are valid and exist in the host).

         Host environment variables are only utilized if the module variables
         ``pyion.ION_PWD`` and ``pyion.ION_NODE_LIST_DIR`` are None.
    """
    # Initialize variables
    ion_nld = ION_NODE_LIST_DIR

    # If not set, use environment variables
    if ion_nld is None:
        ion_nld = os.environ.get('ION_NODE_LIST_DIR')

    # Convert to valid path
    try:
        ion_nld = Path(ion_nld).resolve().absolute()
    except TypeError:
        msg = '\n'.join(_env_var_missing)
        raise OSError(msg.format('ION_NODE_LIST_DIR', ion_nld, 'ION_NODE_LIST_DIR', 'ION_NODE_LIST_DIR'))

    # If path does not exists, raise error
    if not ion_nld.exists():
        raise IOError(_path_missing.format(ion_nld))

    # Set the module variables
    return ion_nld

def set_ion_node_list_dir(path):
    """ Set the ION_NODE_LIST_DIR environment variable. Is not a valid
        OS path, an exception is raise.
        
        :param str: Relative or absolute path.
    """
    # If empty, just return
    if path is None: return

    # Check consistency of this variable
    ION_NODE_LIST_DIR = check_ion_env_vars(path)
    
    # Set it in the environment
    os.environ['ION_NODE_LIST_DIR'] = str(ION_NODE_LIST_DIR)

def _register_proxy(proxy_map, key, proxy_cls, *args, **kwargs):
    """ Implements the singleton mechanism for proxies """
    # If it already exists, return it
    if key in proxy_map: return proxy_map[key]

    # Create a new one. It gets registered during initialization
    proxy = proxy_cls(*args, **kwargs)

    # Register this proxy
    proxy_map[key] = proxy

    # Return new proxy
    return proxy

def _unregister_proxy(proxy_map, node_nbr):
    # Try to unregister node. If not possible, return
    try:
        del proxy_map[str(node_nbr)]
    except KeyError:
        pass

def rel2abs_time(str_time):
    """ Assumes that relative times are provided as ``+xxxx`` where
        xxxx is the time gap in seconds.
    """
    # If this is an absolute time, return
    if '+' not in str_time: return str_time        

    # Compute the date from now
    tstart = datetime.now() + timedelta(seconds=int(str_time[1:]))

    # Transform to ION time format
    return tstart.strftime('%Y/%m/%d-%H:%M:%S')

class Rate():
    def __init__(self, rate):
        self.rate = rate
        self.tic  = time.time()

    def sleep(self):
        # Sleep for a while
        time.sleep(max(0, 1.0/self.rate - (time.time()-self.tic)))

        # Reset tic
        self.tic = time.time()


""" 
# ===========================================================================
# Defines management ION functions.
#
# Author: Marc Sanchez Net
# Date:   08/21/2019
# Copyright (c) 2019, California Institute of Technology ("Caltech").  
# U.S. Government sponsorship acknowledged.
# ===========================================================================
"""

# General imports
from unittest.mock import Mock
from warnings import warn

# pyion imports
from . import utils

# Import C Extension
from . import _mgmt

# Define all methods/vars exposed at pyion
_cgr    = ['cgr_list_contacts', 'cgr_list_ranges', 'cgr_add_contact', 
           'cgr_add_range', 'cgr_delete_contact', 'cgr_delete_range',
           'cgr_list_regions']
_bp     = ['bp_endpoint_exists', 'bp_add_endpoint', 'bp_list_endpoints']
_ltp    = ['ltp_span_exists']
_cfdp   = []
__all__ = ['find_span', 'sm_task_yield'] + _cgr + _bp + _ltp + _cfdp

# ============================================================================
# === Miscellaneous ION functions
# ============================================================================

def find_span(remote_engine_id):
    return _mgmt.find_span(remote_engine_id)

def sm_task_yield():
    _mgmt.sm_task_yield

# ============================================================================
# === Functions to add endpoints
# ============================================================================

def bp_add_endpoint(eid, discard=True):
    """ Define a new endpoint in ION

        :param str eid: E.g., ipn:1.1. (ipn:x.0 not valid)
        :param bool discard: If True, bundles arriving to and endpoint that 
                             has not application attached to it will be discarded.
                             Otherwise they will be enqueued for later delivery.
    """
    # If this endpoint is already defined, skip
    if bp_endpoint_exists(eid): return

    # Add the new endpoint.
    _mgmt.bp_add_endpoint(eid, int(discard))

def bp_endpoint_exists(eid):
    """ Check if an endpoint has been defined in ION.

        :param str eid: E.g., ipn:1.1.
        :return bool: True if endpoint is defined in ION
    """
    return _mgmt.bp_endpoint_exists(eid)

def bp_list_endpoints():
    """ List all endpoints defined in ION """
    raise NotImplementedError

# ============================================================================
# === Functions to manage regions
# ============================================================================

def cgr_list_regions():
    """ List all the regions defined in this ION node

        :return List[int]: List of region numbers
    """
    return _mgmt.list_regions()

# ============================================================================
# === Functions to create/delete/modify the contact plan
# ============================================================================

def cgr_list_contacts():
    """ List all contacts in ION's contact plan.
        
        :return List[Dict]: Each dict is a contact defined as:
                            {orig:, dest:, tstart:, tend:, rate:, confidence:, }

        .. Tip:: The data rate is returned in [bits/sec]
    """
    return _mgmt.list_contacts()

def cgr_list_ranges():
    """ List all ranges in ION's contact_plan 

        :return List[Dict]: Each dict is a range defined as:
                            {orig:, dest:, tstart:, tend:, owlt:}
    """
    return _mgmt.list_ranges()

def cgr_add_contact(orig, dest, tstart, tend, rate, region_nbr=1, confidence=1.0, announce=True):
    """ Add a contact to ION's contact plan

        :param int orig: Node number of the contact origin
        :param int dest: Node number of the contact destination
        :param str tstart: Contact start time. Format is ``yyyy/mm/dd-hh:MM:ss`` or ``+0``
        :param str tend: Contact end time. Format is ``yyyy/mm/dd-hh:MM:ss`` or ``+0``
        :param float rate: Contact data rate in [bits/sec].
        :param int region_nbr: Region index. Defaults to 1.
        :param float confidence: Contact confidence. Defaults to 1. 
        :param bool announce: If True, the information of this contact will be multicasted
                              to all nodes in the region. Default is True

        .. Warning:: The contact data rate gets transformed to bytes/sec 
                     and rounded internally
    """
    # If tstart/tend is/are provided in relative format, transform it to absolute
    tstart = utils.rel2abs_time(tstart)
    tend   = utils.rel2abs_time(tend)

    # Add the contact
    _mgmt.add_contact(region_nbr, orig, dest, tstart, tend, int(rate/8), confidence, int(announce))

def cgr_add_range(orig, dest, tstart, tend, owlt=0.0, announce=True):
    """ Add a range to ION's contact plan

        :param int orig: Node number of the contact origin
        :param int dest: Node number of the contact destination
        :param str tstart: Contact start time. Format is ``yyyy/mm/dd-hh:MM:ss``
        :param str tend: Contact end time. Format is ``yyyy/mm/dd-hh:MM:ss``
        :param float owlt: Range in light seconds.
        :param bool announce: If True, the information of this contact will be multicasted
                              to all nodes in the region. Default is True

        .. Warning:: The owlt gets rounded internally
    """
    # If tstart/tend is/are provided in relative format, transform it to absolute
    tstart = utils.rel2abs_time(tstart)
    tend   = utils.rel2abs_time(tend)

    # Add the range
    _mgmt.add_range(orig, dest, tstart, tend, int(owlt), int(announce))

def cgr_delete_contact(orig, dest, tstart=None, region_nbr=1, announce=True):
    """ Delete a contact from ION's contact plan

        :param int orig: Node number of the contact origin
        :param int dest: Node number of the contact destination. 
        :param str tstart: Contact start time. If None, all contacts between orig and
                           dest are deleted. Format is ``yyyy/mm/dd-hh:MM:ss``
        :param int region_nbr: Region number for this node. Defaults to 1                           
        :param bool announce: If True, the information of this contact will be multicasted
                              to all nodes in the region. Default is True                           
    """
    # If tstart/tend is/are provided in relative format, transform it to absolute
    if tstart: tstart = utils.rel2abs_time(tstart)

    # Delete the contact
    _mgmt.delete_contact(region_nbr, orig, dest, tstart, int(announce))

def cgr_delete_range(orig, dest, tstart=None):
    """ Delete a range from ION's contact plan

        :param int orig: Node number of the contact origin
        :param int dest: Node number of the contact destination
        :param str tstart: Range start time. If None, all ranges between orig and dest
                           are deleted. Format is ``yyyy/mm/dd-hh:MM:ss``
    """
    # If tstart/tend is/are provided in relative format, transform it to absolute
    if tstart: tstart = utils.rel2abs_time(tstart)

    # Delete the range
    _mgmt.delete_range(orig, dest, tstart)

# ============================================================================
# === LTP-related functions
# ============================================================================

def ltp_span_exists(engine_nbr):
    """ Determines if an LTP span is defined within ION.

        :param int engine_nbr: Peer engine number.
        :return bool: True if span exists.
    """
    return _mgmt.ltp_span_exists(engine_nbr)



"""
EmION Plugins — API-based plugin system.

Users provide their own anomaly detection module as a FastAPI endpoint.
EmION calls the user's API during simulation for each node.

See api_template.py for the expected format.
"""

from .base import EmionPlugin



""" 
# ===========================================================================
# Interface between the ION and Python. Provides ``Proxy`` object to connect
# to ION.
#
# Author: Marc Sanchez Net
# Date:   04/17/2019
# Copyright (c) 2019, California Institute of Technology ("Caltech").  
# U.S. Government sponsorship acknowledged.
# ===========================================================================
"""

import os
import signal
from pathlib import Path
from time import sleep
# General imports
from unittest.mock import Mock
from warnings import warn

# Module imports
from . import bp
from . import cfdp
from . import constants as cst
from . import ltp
from . import mem
from . import utils

# Import C Extensions
# These will be emion.pyion._bp etc.
from . import _bp
from . import _cfdp
from . import _ltp

# Define all methods/vars exposed at pyion
__all__ = ['get_bp_proxy', 'get_cfdp_proxy', 'get_ltp_proxy', 'get_sdr_proxy',
           'get_psm_proxy']

# ============================================================================
# === Singleton of proxies
# ============================================================================

# Map of proxies to BP defined as {node_nbr: BpProxy}
_bp_proxies = {}

# Map of proxies to CFDP defined as {peer_entity_nbr: CfdpProxy}
_cfdp_proxies = {}

# Map of proxies to LTP defined as {client_id: LtpProxy}
_ltp_proxies = {}

# Map of proxies to the SDR defined as {(node_nbr, sdr_name): SdrProxy}
_sdr_proxies = {}

# Map of proxies to the PSM defined as {(node_nbr, wm_key): PsmProxy}
_psm_proxies = {}

def get_bp_proxy(node_nbr):
    """ Returns a BpProxy for a given node number. If it already exists, it
        gives you the already instantiated copy.

        :param:  Node number
        :return: BpProxy object
    """
    global _bp_proxies
    proxy = utils._register_proxy(_bp_proxies, str(node_nbr), BpProxy, node_nbr)
    proxy.bp_attach()
    return proxy

def get_cfdp_proxy(peer_entity_nbr):
    """ Returns a CfdpProxy for a given peer entity number. If it already exists, it
        gives you the already instantiated copy.

        :param:  Peer entity number
        :return: CfdpProxy object
    """
    global _cfdp_proxies
    proxy = utils._register_proxy(_cfdp_proxies, str(peer_entity_nbr), CfdpProxy, peer_entity_nbr)
    proxy.cfdp_attach()
    return proxy

def get_ltp_proxy(client_id):
    """ Returns an LtpProxy for a given client application. If it already exists, it
        gives you the already instantiated copy.

        :param:  Client id (number)
        :return: LtpProxy object
    """
    global _ltp_proxies
    proxy = utils._register_proxy(_ltp_proxies, str(client_id), LtpProxy, client_id)
    proxy.ltp_attach()
    return proxy

def get_sdr_proxy(node_nbr):
    """ Return an SdrProxy for a given client application. If it already exists, it
        gives you th ealready instantiated copy

        :param: Node number
        :return: SdrProxy object
    """
    global _sdr_proxies
    return utils._register_proxy(_sdr_proxies, node_nbr, mem.SdrProxy, node_nbr)

def get_psm_proxy(node_nbr):
    """ Return a PsmProxy for a given client application. If it already exists, it
        gives you th ealready instantiated copy

        :param node_nbr: Node number
        :return: PsmProxy object
    """
    global _psm_proxies
    return utils._register_proxy(_psm_proxies, node_nbr, mem.PsmProxy, node_nbr)

def shutdown():
    """ Shutdowns pyion: All endpoints, access points, etc. """
    # Iterate through all BP endpoints and close them
    print('Closing all BP endpoints... ', end='')
    for proxy in _bp_proxies.values():
        proxy.bp_close_all()
    print('DONE')

    # Iterate through all CFDP entities and close them
    print('Closing all CFDP entities... ', end='')
    for proxy in _cfdp_proxies.values():
        proxy.cfdp_cancel_all()
        proxy.cfdp_close_all()
    print('DONE')

    # Iterate through all LTP access points and close them
    print('Closing all LTP access points...', end='')
    for proxy in _ltp_proxies.values():
        proxy.ltp_interrupt_all()
        proxy.ltp_close_all()
    print('DONE')

    # Iterate through all SDR proxies and close them
    print('Closing all SDR proxies...', end='')
    for proxy in _sdr_proxies.values():
        proxy.close()
    print('DONE')

    # Iterate through all PSM proxies and close them
    print('Closing all PSM proxies...', end='')
    for proxy in _psm_proxies.values():
        proxy.close()
    print('DONE')

# ============================================================================
# === Capture the SIGNIT and trigger cleanup
# ============================================================================

def pyion_sigint_handler(sig, frame):
    """ Default SIGINT handler. It closes all endpoints for all proxies and
        then detaches them from ION.
    """
    print('[pyion] KeyboardInterruption... ')
    shutdown()
    
# Register the signal handler function. If another one was provided previously,
# do not override it, combine it with pyion's
prev_sigint_handler = signal.getsignal(signal.SIGINT)
if prev_sigint_handler is None:
    signal.signal(signal.SIGINT, pyion_sigint_handler)
else:
    def combined_sigint_handler(sig, frame):
        pyion_sigint_handler(sig, frame)
        prev_sigint_handler(sig, frame)
    signal.signal(signal.SIGINT, combined_sigint_handler)
    
# ============================================================================
# === Proxy to BP in ION for a given node
# ============================================================================

class BpProxy(utils.Proxy):
    """ Proxy to the ION enginer running in this host. Do not instantiate 
        manually, use ``pyion.get_bp_proxy`` instead.

        Private Variables
        -----------------
        :ivar _ept_map: {'ipn:1.1': Endpoint_1, 'ipn:1.2': Endpoint_2, ...}
    """
    def __init__(self, node_nbr):
        """ Initializes the proxy, checks that the ION environment variables are
            valid, and attaches to the ION engine.

            :param int node_nbr: Node number for this proxy
        """
        global _bp_proxies

        # If a proxy for this node already exists, raise error
        if node_nbr in _bp_proxies:
            raise KeyError('A proxy for node {} already exists. Use ``pyion.get_bp_proxy`` \
                            to retrieve it'.format(node_nbr))

        # Call parent constructor
        super().__init__(node_nbr)      

        # Map {eid: Endpoint}
        self._ept_map = {}

    def __del__(self):
        """ Close all Endpoints associated with this proxy """
        global _bp_proxies
        self.bp_close_all()
        self.bp_detach()
        utils._unregister_proxy(_bp_proxies, self.node_nbr)

    def is_endpoint_open(self, eid):
        """ Check if this Proxy thinks a given EID is currently opened
            
            :param EID: EID
            :return: True if it is open
        """
        return eid in self._ept_map

    @property
    def open_endpoints(self):
        """ List all EIDs that are this Proxy sees open

            :return: Tuple of EIDS
        """
        return tuple(getattr(self, '_ept_map', {}).values())

    @utils.in_ion_folder
    def bp_attach(self):
        """ Attach to ION """
        # Attach to ION instance
        _bp.bp_attach()

        # Mark as attached to ION
        self.attached = True

    @utils.in_ion_folder
    def bp_detach(self):
        """ Dettach from ION """
        # Detach from ION instance
        _bp.bp_detach()

        # Mark as detached from ION
        self.attached = False

    @utils._chk_attached
    @utils.in_ion_folder
    def bp_open(self, eid, TTL=3600, priority=cst.BpPriorityEnum.BP_STD_PRIORITY,
                report_eid=None, custody=cst.BpCustodyEnum.NO_CUSTODY_REQUESTED,
                report_flags=cst.BpReportsEnum.BP_NO_RPTS, ack_req=cst.BpAckReqEnum.BP_NO_ACK_REQ,
                retx_timer=0, chunk_size=None, timeout=None, mem_ctrl=False):
        """ Open an endpoint. If it already exists, the existing instance
            is returned.

            .. Tip:: The parameters specified when opening the endpoint are
                     used as default values. When calling ``bp_send``, they
                     can be overriden.
            .. Tip:: To see longer explanation for these parameters, see
                     ``man bp`` in shell terminal

            :param EID: ID of the enpoint being opened
            :param TTL: Time-to-live [sec] for all bundles through this endpoint. 
                        Default is 3600 seconds
            :param priority: Class of service for all bundles through this endpoint.
                             Default is ``BpPriorityEnum.BP_STD_PRIORITY``
            :param report_eid: Endpoint id where report bundles will be sent to
            :param custody: Type of custody required for the bundles issued by this
                            endpoint. Default is ``BpCustodyEnum.NO_CUSTODY_REQUIRED``
            :param report_flags: Flag indicating which reports to request for a bundle.
                                 Default is ``BpReportsEnum.BP_NO_RPTS``. Example
                                 ``BpReportsEnum.BP_RECEIVED_RPT | BpReportsEnum.BP_CUSTODY_RPT``.
            :param ack_req: If True, it indicates that the application on top of BP is 
                            expecting to receive and end-to-end acknowledgement. Default
                            is ``BpAckReqEnum.BP_NO_ACK_REQ``
            :param retx_timer: Custodial timer retransmission in [sec]. Defaults to 0, which 
                               means that no timer is created.
            :param chunk_size: Send/Receive data in bundles of ``chunk_size`` bytes (plus header), 
                               instead of a single potentially very large bundle.
            :param timeout: If specified, the endpoint will be interrupted if timeout
                            seconds elapse without receiving anything.
            :return: Endpoint object
        """
        # If this EID is already open, return it
        if self.is_endpoint_open(eid):
            return self._ept_map[eid]

        # Detect if this endpoint "detains" bundles (see bp_open_source in ION manual)
        detained = (retx_timer > 0)

        # Open EID in ION. Get the address of the BpSapState as a long. If retx_timer>0, then
        # open the endpoint in "detained mode" (see bp_open_source in ION manual).
        sap_addr = _bp.bp_open(eid, int(detained), int(mem_ctrl))

        # Create an endpoint
        ept_obj = bp.Endpoint(self, eid, sap_addr, TTL, int(priority), report_eid,
                              int(custody), int(report_flags), int(ack_req), 
                              int(retx_timer), detained, chunk_size, timeout, mem_ctrl)

        # Store it
        self._ept_map[eid] = ept_obj

        # Return it
        return ept_obj

    @utils._chk_attached
    @utils.in_ion_folder
    def bp_close(self, ept):
        """ Close an endpoint. If it is not open, ``ConnectionError`` is raised.

            :param: Endpoint object to close
        """
        # If object passed is not endpoint, fail
        if not isinstance(ept, bp.Endpoint):
            raise ValueError('Expected an Enpoint instance, '+ str(type(ept)))

        # Interrupt this endpoint first. 
        try:
            self.bp_interrupt(ept)
        except ConnectionError:
            raise ConnectionError('Cannot close endpoint {}. It is not open.'.format(ept.eid))

        # Get SAP address in memory.
        ept_obj = self._ept_map.pop(ept.eid)

        # If already closed, you are done
        if not ept_obj.is_open:
            return

        # Close EID in ION
        _bp.bp_close(ept_obj._sap_addr)

        # Mark object as inactive
        ept_obj._cleanup()

    @utils._chk_attached
    @utils.in_ion_folder
    def bp_interrupt(self, ept):
        """ Interrupt and endpoint while receiving data. If it is not open, ``ConnectionError`` is raised.
            If this endpoint is not blocked receiving, this call has no effect.

            :param: Endpoint object to close
        """
        # If object passed is not endpoint, fail
        if not isinstance(ept, bp.Endpoint):
            raise ValueError('Expected an Enpoint instance')
        
        # If this EID is not open, return
        if not self.is_endpoint_open(ept.eid):
            raise ConnectionError('Cannot interrupt endpoint {}. It is not open.'.format(ept.eid))

        # Interrupt endpoint in ION
        _bp.bp_interrupt(ept._sap_addr)

        # This is little hack to ensure that SIGINT prints "interrupted" as opposed
        # to raising ConnectionAbortedError. However, this is not guaranteed
        sleep(0.001)

    def bp_close_all(self):
        """ Close all opened endpoints """
        # Close all endpoints
        for ept in self.open_endpoints:
            self.bp_close(ept)

    def bp_interrupt_all(self):
        """ Interrupt all opened endpoints """
        # Close all endpoints
        for ept in self.open_endpoints:
            self.bp_interrupt(ept)

# ============================================================================
# === Proxy to CFDP engine in ION for a given node
# ============================================================================

class CfdpProxy(utils.Proxy):
    def __init__(self, node_nbr):
        # Call parent constructor
        super().__init__(node_nbr)

        # Map {entity_nbr: Entity}
        self._ett_map = {}

    def __del__(self):
        """ Close all Endpoints associated with this proxy """
        global _cfdp_proxies
        self.cfdp_close_all()
        self.cfdp_detach()  
        utils._unregister_proxy(_cfdp_proxies, self.node_nbr)    

    def is_entity_open(self, peer_entity_nbr):
        """ Check if this CfdpProxy thinks a given Entity is currently opened
            
            :param EID: EID
            :return: True if it is open
        """
        return peer_entity_nbr in self._ett_map

    @property
    def open_entities(self):
        """ List all EIDs that are this Proxy sees open

            :return: Tuple of EIDS
        """
        return tuple(getattr(self, '_ett_map', {}).keys())

    @utils.in_ion_folder
    def cfdp_attach(self):
        """ Attach to ION's CFDP engine """
        # Attach to ION instance
        _cfdp.cfdp_attach()

        # Mark as attached to ION
        self.attached = True

    @utils.in_ion_folder
    def cfdp_detach(self):
        """ Dettach from ION """
        # Detach from ION instance
        _cfdp.cfdp_detach()

        # Mark as detached from ION
        self.attached = False

    @utils._chk_attached
    @utils.in_ion_folder
    def cfdp_open(self, peer_entity_nbr, endpoint, mode=cst.CfdpMode.CFDP_BP_RELIABLE,
                  closure_latency=cst.CfdpClosure.CFDP_NO_CLOSURE, 
                  seg_metadata=cst.CfdpMetadataEnum.CFDP_NO_SEG_METADATA):
        """ Open a CFDP entity

            :param peer_entity_nbr:
            :param endpoint: Enpoint object through which the CFDP will run
            :param mode: Reliable vs. Non-reliable. See all options in ``pyion.constants``
            :param closure_latency: See ``help(CfdpClosure)``
            :param seg_metadata:
        """ 
        # If already open, return it
        if peer_entity_nbr in self._ett_map:
            return self._ett_map[peer_entity_nbr]

        # Open a proxy to the CFDP engine
        param_addr = _cfdp.cfdp_open(
            peer_entity_nbr, 
            endpoint.TTL,
            endpoint.priority,
            endpoint.sub_priority,
            endpoint.report_flags,
            endpoint.criticality,
        )

        # Create a CFDP entity object
        ett_obj = cfdp.Entity(self, peer_entity_nbr, param_addr, endpoint,
                              int(mode), int(closure_latency), int(seg_metadata))

        # Store it
        self._ett_map[peer_entity_nbr] = ett_obj

        # Return it
        return ett_obj

    @utils._chk_attached
    @utils.in_ion_folder
    def cfdp_close(self, peer_entity_nbr):
        """ Close a CFDP entity

            :param peer_entity_nbr:
        """ 
        # Get SAP address in memory.
        ett_obj = self._ett_map.pop(peer_entity_nbr)

        # If already closed, you are done
        if not ett_obj.is_open:
            return

        # Close Entity in ION
        _cfdp.cfdp_close(ett_obj._param_addr)

        # Mark object as inactive
        ett_obj._cleanup()

    def cfdp_close_all(self):
        """ Close all open entities in this node """
        for peer_entity_nbr in self.open_entities:
            self.cfdp_close(peer_entity_nbr)

    def cfdp_cancel_all(self):
        """ Cancel any transaction in all entities in this node """
        for peer_nbr in self.open_entities:
            self._ett_map[peer_nbr].cfdp_cancel()

# ============================================================================
# === Proxy to LTP engine in ION for a given node
# ============================================================================

class LtpProxy(utils.Proxy):
    def __init__(self, node_nbr):
        # Call parent constructor
        super().__init__(node_nbr)

        # Map {cliend id: AccessPoint}
        self._sap_map = {}

    def __del__(self):
        """ Close all access points associated with this proxy """
        global _ltp_proxies
        self.ltp_close_all()
        self.ltp_detach()
        utils._unregister_proxy(_ltp_proxies, self.node_nbr)

    def is_client_open(self, client_id):
        """ Check if this LtpProxy thinks a given client is already using the
            LTP engine
            
            :param client_id: The id of the client application (int)
            :return: True if it is open
        """
        return client_id in self._sap_map

    @property
    def open_clients(self):
        """ List all client ids that are this Proxy sees open

            :return: Tuple of client ids
        """
        return tuple(getattr(self, '_sap_map', {}).keys())

    @utils.in_ion_folder
    def ltp_attach(self):
        """ Attach to ION's LTP engine """
        # Attach to ION instance
        _ltp.ltp_attach()

        # Mark as attached to ION
        self.attached = True

    @utils.in_ion_folder
    def ltp_detach(self):
        """ Dettach from ION """
        # Detach from ION's LTP engine
        _ltp.ltp_detach()

        # Mark as detached from ION
        self.attached = False

    @utils._chk_attached
    @utils.in_ion_folder
    def ltp_open(self, client_id):
        """ Open a service access point for a client 

            :param client_id:
        """ 
        # If already open, return it
        if client_id in self._sap_map:
            return self._sap_map[client_id]

        # Open ltp access point
        sap_addr = _ltp.ltp_open(client_id)

        # Create an LTP Access Point
        sap_obj = ltp.AccessPoint(self, client_id, sap_addr)

        # Store it
        self._sap_map[client_id] = sap_obj

        # Return it
        return sap_obj

    @utils._chk_attached
    @utils.in_ion_folder
    def ltp_close(self, client_id):
        """ Close an LTP access point

            :param client_id:
        """
        # Get SAP address in memory.
        sap_obj = self._sap_map.pop(client_id)

        # If already closed, you are done
        if not sap_obj.is_open:
            return

        # Close Entity in ION
        _ltp.ltp_close(sap_obj._sap_addr)

        # Mark object as inactive
        sap_obj._cleanup()

    def ltp_close_all(self):
        """ Close all open clients in this proxy """
        for client_id in self.open_clients:
            self.ltp_close(client_id)

    def ltp_interrupt_all(self):
        """ Interrupt any LTP transactions in all clients in this proxy """
        for client_id in self.open_clients:
            self._sap_map[client_id].ltp_interrupt()

# ============================================================================
# === EOF
# ============================================================================



""" 
# ===========================================================================
# Defines proxies to ION's SDR and PSM so that you can monitor them.
#
# Author: Marc Sanchez Net
# Date:   08/12/2019
# Copyright (c) 2019, California Institute of Technology ("Caltech").  
# U.S. Government sponsorship acknowledged.
# ===========================================================================
"""

# Generic imports
import abc
from collections import defaultdict
from pprint import pprint
from unittest.mock import Mock
from threading import Thread
import time
from warnings import warn

# Import pyion modules
from . import utils

# Import C Extension
from . import _mem

# Define all methods/vars exposed at pyion
__all__ = ['SdrProxy', 'PsmProxy']

# ============================================================================
# === Abstract Memory Proxy
# ============================================================================

class MemoryProxy(utils.Proxy, abc.ABC):
    """ Abstract class parent to SdrProxy and PsmProxy """
    def __init__(self, node_nbr):
        # Call parent constructor
        super().__init__(node_nbr)

        # Variables for monitor
        self._th         = None
        self._monitor_on = False
        self._rate       = None
        self._tspan      = -1
        self._print_res  = False
        self._results    = defaultdict(dict)

    @abc.abstractmethod
    def dump(self):
        """ Dump memory state """
        pass

    def close(self):
        # If no monitoring thread, return
        if self._th is None: return

        # If monitoring thread is not active, return
        if not self._th.is_alive(): return
        
        # Wait for thread to stop
        self._monitor_on = False
        self._th.join()

        # Delete results
        del self._results

    def clean_results(self):
        """ Clean all stored results """
        self._results = defaultdict(dict)

    def start_monitoring(self, rate=1, timespan=-1, print_results=False):
        """ Start monitoring the SDR for a while
            
            :param int rate: Rate in [Hz]
            :param float timespan: Span of time during which measurements
                                   are taken. After that, the measurements
                                   are deleted and new data is acquired.
                                   The default value of -1 indicates that
                                   measurements are never deleted.
            :param float print_results: If True, the state of the memory is printed
                                        to stdout prior to being deleted
        """
        # If a monitoring session is already on, error
        if self._monitor_on:
            raise ValueError('Memory is already being monitored.')

        # Prepare monitoring session
        self._rate       = utils.Rate(rate)
        self._monitor_on = True
        self._tspan      = timespan
        self._print_res  = print_results

        # Start monitor in separate thread
        self._th = Thread(target=self._start_monitoring, daemon=True)
        self._th.start()

    def _start_monitoring(self):
        # Initialize variables
        tic = time.time()
        summary = {}

        while self._monitor_on:
            # Measurement index
            t = time.time()

            # If enough time has elapsed, clear the results
            if self._tspan > 0 and time.time()-tic >= self._tspan:
                # Print if necessary
                if self._print_res: 
                    print('Num. measurements:', len(self._results['summary']))
                    pprint(summary)

                # Clean results and restart time counter
                self.clean_results()
                tic = time.time()

            # Take measurement
            try:
                summary, small_pool, large_pool = self.dump()
            except Exception as e:
                summary, small_pool, large_pool = e, None, None

            # Store results
            self._results['summary'][t]    = summary
            self._results['small_pool'][t] = small_pool
            self._results['large_pool'][t] = large_pool

            # Sleep for a while
            if self._rate: self._rate.sleep()

    def stop_monitoring(self):
        """ Stop monitoring the SDR and return results"""
        # If no monitoring session is on, return
        if not self._monitor_on: return

        # Signal exit to the monitor thread
        self._monitor_on = False
        self._rate = None
        self._tspan = -1

        # Wait for monitoring thread to exit
        self._th.join()

        # Return collected results
        return self._results

    def __str__(self):
        return '<{}>'.format(self.__class__.__name__)

# ============================================================================
# === SDR PROXY
# ============================================================================

class SdrProxy(MemoryProxy):
    """ Proxy object to ION's SDR 

        :ivar int: Node number
        :ivar str: SDR name (see ``ionconfig``, option ``sdrName``)
    """
    @utils.in_ion_folder
    def dump(self):
        """ Dump the current state of the SDR 
        
        :return Tuple[Dict]: usage summary, small pool free block count,
                             large pool free block count
        """
        return _mem.sdr_dump()

# ============================================================================
# === PSM PROXY
# ============================================================================

class PsmProxy(MemoryProxy):
    """ Proxy object to ION's PSM 
    
        :ivar int wm_key: See ``wmKey`` in ``ionconfig``. Defaults to 65281 
                          (i.e., ION's default value)
        :ivar int wm_size: Unused
        :ivar str partition_name: Unused
    """
    @utils.in_ion_folder
    def dump(self):
        """ Dump the current state of the PSM
        
            :return Tuple[Dict]: usage summary, small pool free block count,
                                 large pool free block count
        """
        return _mem.psm_dump()


""" 
# ===========================================================================
# Interface between the ION's implementation of the CCSDS CFDP and Python. 
# Internally, all classes call the C Extension _cfdp.
#
# Author: Marc Sanchez Net
# Date:   04/17/2019
# Copyright (c) 2019, California Institute of Technology ("Caltech").  
# U.S. Government sponsorship acknowledged.
# ===========================================================================
"""

# General imports
from unittest.mock import Mock
from pathlib import Path
from threading import Event, Thread
from warnings import warn

# Module imports
from . import utils
from .constants import CfdpEventEnum

# Import C Extension
from . import _cfdp

# Define all methods/vars exposed at pyion
__all__ = ['Entity']

# ============================================================================
# === Entity class
# ============================================================================

class Entity():
	""" Class to represent a CFDP entity. """
	def __init__(self, proxy, peer_entity_nbr, param_addr, endpoint, mode,
				 closure_latency, seg_metadata):
		""" Class initializer """
		# Store variables
		self.proxy        = proxy
		self.entity_nbr   = peer_entity_nbr
		self.node_dir     = proxy.node_dir
		self._param_addr  = param_addr
		self.endpoint     = endpoint
		self.mode         = mode
		self.closure_lat  = closure_latency
		self.seg_metadata = seg_metadata

		# Map {event_id}: event handler function
		self.event_handers = {}

		# Create a condition to signal end of transaction
		self._end_transaction = Event()
		self._ok_transaction = False

		# Start a thread to monitor all events
		self.th = Thread(target=self._monitor_events, daemon=False)
		self.th.start()

	def __del__(self):
		# If you have already been closed, return
		if not self.is_open:
			return

		# Close the Entity and free C memory
		self.proxy.cfdp_close(self.entity_nbr)

	@property
	def is_open(self):
		""" Returns True if entity is opened """
		return (self.proxy is not None and self._param_addr is not None)

	def _cleanup(self):
		""" Clean entity after closing. Do not call directly,
			use ``proxy.cfdp_close``
		"""
		# Clear variables
		self.proxy       = None
		self.entity_nbr  = None
		self._param_addr = None
		self.endpoint    = None

		# Stop the thread monitoring CFDP events
		_cfdp.cfdp_interrupt_events()

		# Wait until the event thread has finished
		self.th.join()

		# Mark end of transactions to wake up threads
		self._mark_transaction_end(False)
		
	@utils._chk_is_open
	@utils.in_ion_folder
	def cfdp_send(self, source_file, dest_file=None, mode=None,
				  closure_lat=None, seg_metadata=None):
		""" Send a file using CFDP to the peer engine for this entity

			:param source_file: str or Path of file to send
			:param dest_file: str or Path. Name of file at receiving
							  engine. It defaults to source_file
			:param **kwargs: See ``proxy.cfdp_send``
		"""
		# Create path object
		src_file = Path(source_file)

		# Check that the source file exists
		if not src_file.exists():
			raise FileNotFoundError('Source file {} does not exist'.format(src_file))

		# Initialize variables
		src_file = src_file.resolve().absolute()

		# Set default values
		dst_file = source_file if dest_file is None else dest_file
		
		# Set default values if necessary
		if mode is None: mode = self.mode
		if closure_lat is None: closure_lat = self.closure_lat
		if seg_metadata is None: seg_metadata = self.seg_metadata

		# Mark that the current transaction has not succeeded yet
		self._ok_transaction = False

		# Trigger CFDP send
		_cfdp.cfdp_send(self._param_addr, str(src_file), str(dst_file), 
							  closure_lat, seg_metadata, mode)

	@utils._chk_is_open
	@utils.in_ion_folder
	def cfdp_request(self, source_file, dest_file=None, mode=None,
				  	 closure_lat=None, seg_metadata=None):
		""" Request a file to be sent to this node using CFDP

			:param source_file: str or Path of file to request
			:param dest_file: str or Path. Name of file at this node 
							  once it is received. Defaults to ``source_file``
			:param **kwargs: See ``proxy.cfdp_send``
		"""
		# Set default values if necessary
		if mode is None: mode = self.mode
		if closure_lat is None: closure_lat = self.closure_lat
		if seg_metadata is None: seg_metadata = self.seg_metadata

		# Mark that the current transaction has not succeeded yet
		self._ok_transaction = False

		# Trigger CFDP request
		_cfdp.cfdp_request(self._param_addr, str(source_file), str(dest_file), 
								 closure_lat, seg_metadata, mode)

	@utils._chk_is_open
	@utils.in_ion_folder
	def cfdp_cancel(self):
		""" Cancel the current CFDP transaction """
		_cfdp.cfdp_cancel(self._param_addr)

	@utils._chk_is_open
	@utils.in_ion_folder
	def cfdp_suspend(self):
		""" Suspend the current CFDP transaction """
		_cfdp.cfdp_suspend(self._param_addr)

	@utils._chk_is_open
	@utils.in_ion_folder
	def cfdp_resume(self):
		""" Resume the current CFDP transaction """
		_cfdp.cfdp_resume(self._param_addr)

	@utils._chk_is_open
	@utils.in_ion_folder
	def cfdp_report(self):
		""" Request issuance on the transmission/reception progress of the current CFDP
			transaction.
		"""
		_cfdp.cfdp_report(self._param_addr)

	@utils._chk_is_open
	def add_usr_message(self, msg):
		""" Add a user message to all CFDP PDUs in the next transaction 

			:param str: User message to add
		"""
		_cfdp.cfdp_add_usr_msg(self._param_addr, msg)

	@utils._chk_is_open
	def add_filestore_request(self, action, file1, file2=None):
		""" Add a user message to all CFDP PDUs in the next transaction 

			:param action: See ``pyion.CFDP_CREATE_FILE``, etc.
			:param file1: String or Path-object
			:param file2: None, string or Path-object.
		"""
		# Transform files to string
		file1 = str(file1)
		if file2 is not None: file2 = str(file2)

		# Call pyion function
		_cfdp.cfdp_add_filestore_request(self._param_addr, action, file1, file2)
	
	def register_event_handler(self, event, func):
		""" Register and event handler for this entity

			:param event: See ``pyion.constants.CFDP_CREATE_FILE``, etc.
			:param func: Function handle with the following signature
						 ``def ev_handler(params)``, where params is a 
						 dictionary with contents that depend on the type
						 of event (see CCSDS CFDP spec, section 3.5.6)
		"""
		self.event_handers[event] = func

	def wait_for_transaction_end(self, timeout=None):
		""" Blocks the calling thread until the transaction has
			finished or the timeout expires.

			:param timeout: Time to wait in [seconds]
			:return: True if transaction finished successfully
		"""
		# Wait for the end of the transaction
		self._end_transaction.wait(timeout=timeout)

		# Return state of transaction
		return self._ok_transaction

	def _mark_transaction_end(self, success):
		""" Mark that the current CFDP transaction has ended 
		
			:param success: True/False
		"""
		# Mark the transaction as successful
		self._ok_transaction = success

		# Awake all threads that were waiting
		self._end_transaction.set()

		# Reset the event
		self._end_transaction.clear()

	def _monitor_events(self):
		""" Monitor all CFDP events """
		while self.is_open:
			# Get the next event
			evt, ev_params = _cfdp.cfdp_next_event()

			# Create event type class from integer code
			evt = CfdpEventEnum(evt)

			# If no event, continue
			if evt == CfdpEventEnum.CFDP_NO_EVENT:
				continue

			# If you have an event handler for this event type, use it
			try:
				func = self.event_handers[evt]
				func(evt, ev_params)
			except KeyError:
				pass

			# If you have a handler for all events, use it
			try:
				func = self.event_handers[CfdpEventEnum.CFDP_ALL_IND]
				func(evt, ev_params)
			except KeyError:
				pass

			# If transaction finished ok, report it
			if evt == CfdpEventEnum.CFDP_TRANSACTION_FINISHED_IND:
				self._mark_transaction_end(True)

			# If transaction failed, report it
			if evt == CfdpEventEnum.CFDP_ABANDONED_IND:
				self._mark_transaction_end(False)
	
	def __enter__(self):
		""" Allows an endpoint to be used as context manager """
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		""" Allows an endpoint to be used as context manager """
		pass

	def __str__(self):
		return '<Entity: {} ({})>'.format(self.entity_nbr, 'Open' if self.is_open else 'Closed')

	def __repr__(self):
		return '<Entity: {} ({})>'.format(self.entity_nbr, self._param_addr)


""" 
# ===========================================================================
# This file constains all the constants used in pyion
#
# Author: Marc Sanchez Net
# Date:   04/17/2019
# Copyright (c) 2019, California Institute of Technology ("Caltech").  
# U.S. Government sponsorship acknowledged.
# ===========================================================================
"""

# Generic imports
from enum import IntEnum, unique
from unittest.mock import Mock
from warnings import warn

# Import C Extensions
import _bp
import _cfdp

# Define all methods/vars exposed at pyion
__all__ = [
    'BpCustodyEnum', 
    'BpPriorityEnum',
    'BpEcsEnumeration',
    'BpReportsEnum',
    'BpAckReqEnum',
    'CfdpMode',
    'CfdpClosure',
    'CfdpMetadataEnum',
    'CfdpFileStoreEnum',
    'CfdpEventEnum',
    'CfdpConditionEnum',
    'CfdpFileStatusEnum',
    'CfdpDeliverCodeEnum'
]

# ============================================================================
# === BUNDLE PROTOCOL
# ============================================================================

@unique
class BpCustodyEnum(IntEnum):
    """ BP custody switch enumeration. See ``help(CustodyEnum)`` """
    NO_CUSTODY_REQUESTED    = _bp.NoCustodyRequested
    SOURCE_CUSTODY_OPTIONAL = _bp.SourceCustodyOptional
    SOURCE_CUSTODY_REQUIRED = _bp.SourceCustodyRequired

@unique
class BpPriorityEnum(IntEnum):
    """ BP priority enumeration. See ``help(BpPriorityEnum)`` """
    BP_BULK_PRIORITY      = _bp.BP_BULK_PRIORITY
    BP_STD_PRIORITY       = _bp.BP_STD_PRIORITY
    BP_EXPEDITED_PRIORITY = _bp.BP_EXPEDITED_PRIORITY

@unique
class BpEcsEnumeration(IntEnum):
    """ BP extended class-of-service enumeration. See ``help(BpEcsEnumeration)`` 
    
        - BP_MINIMUM_LATENCY: Forward bundle on all routes
        - BP_BEST_EFFORT: Send using an unreliable convergence layer
        - BP_RELIABLE: Send using a reliable convergence layer
        - BP_RELIABLE_STREAMING: BP_BEST_EFFORT | BP_RELIABLE
    """
    BP_MINIMUM_LATENCY    = _bp.BP_MINIMUM_LATENCY  
    BP_BEST_EFFORT        = _bp.BP_BEST_EFFORT      
    BP_RELIABLE           = _bp.BP_RELIABLE         
    BP_RELIABLE_STREAMING = getattr(_bp, 'BP_RELIABLE_STREAMING', 99)

@unique
class BpReportsEnum(IntEnum):
    """ BP reports enumeration. See ``help(BpReportsEnum)`` """
    BP_NO_RPTS       = 0
    BP_RECEIVED_RPT  = _bp.BP_RECEIVED_RPT
    BP_CUSTODY_RPT   = _bp.BP_CUSTODY_RPT
    BP_FORWARDED_RPT = _bp.BP_FORWARDED_RPT
    BP_DELIVERED_RPT = _bp.BP_DELIVERED_RPT
    BP_DELETED_RPT   = _bp.BP_DELETED_RPT

@unique
class BpAckReqEnum(IntEnum):
    """ BP application acknowledgement enumeration. See ``help(BpAckReqEnum)`` """
    BP_ACK_REQ    = True
    BP_NO_ACK_REQ = False

# ============================================================================
# === CFDP PROTOCOL
# ============================================================================

@unique
class CfdpMode(IntEnum):
    """ CFDP mode enumeration. See ``help(CfdpMode)`` 
    
        - CFDP_CL_RELIABLE: Reliability provided by the convergence layer
        - CFDP_UNRELIABLE: No reliability provided
        - CFDP_BP_RELIABLE: Reliability provided by BP custody transfer
    """
    CFDP_CL_RELIABLE = 0    # Reliability provided by convergence layer
    CFDP_UNRELIABLE  = 1    # No reliability provided
    CFDP_BP_RELIABLE = 2    # Reliability provided by custody transfer

@unique
class CfdpClosure(IntEnum):
    """ Length of time following transmission of the EOF PDU within which a responding
        Transaction Finish PDU is expected. If no Finish PDU is requested, this parameter value should be
        zero.

        .. Tip:: If you need a number of seconds not specified as an option in this enumeration,
                 simply pass it directly.
    """
    CFDP_NO_CLOSURE = 0     # No closure.
    CFDP_1_SEC      = 1
    CFDP_1_MIN      = 60
    CFDP_5_MIN      = 300
    CFDP_1_HR       = 3600

@unique
class CfdpMetadataEnum(IntEnum):
    """ CFDP segment metadata options. See ``help(CfdpMetadataEnum)``

        - CFDP_NO_SEG_METADATA: No metadata added
        - CFDP_SEG_METADATA: Adds current time as string to all PDUs
    """
    CFDP_NO_SEG_METADATA = 0
    CFDP_SEG_METADATA    = 1

@unique
class CfdpFileStoreEnum(IntEnum):
    """ CFDP file store actions. See ``help(CfdpFileStoreEnum)`` """
    CFDP_CREATE_FILE  = _cfdp.CfdpCreateFile
    CFDP_DELETE_FILE  = _cfdp.CfdpDeleteFile
    CFDP_RENAME_FILE  = _cfdp.CfdpRenameFile
    CFDP_APPEND_FILE  = _cfdp.CfdpAppendFile
    CFDP_REPLACE_FILE = _cfdp.CfdpReplaceFile
    CFPD_CREATE_DIR   = _cfdp.CfdpCreateDirectory
    CFDP_REMOVE_DIR   = _cfdp.CfdpRemoveDirectory
    CFDP_DENY_FILE    = _cfdp.CfdpDenyFile
    CFDP_DENY_DIR     = _cfdp.CfdpDenyDirectory

@unique
class CfdpEventEnum(IntEnum):
    """ CFDP event types. See ``help(CfdpEventEnum)`` and Section 3.5.6 of CCSDS CFDP """
    CFDP_NO_EVENT                 = _cfdp.CfdpNoEvent
    CFDP_TRANSACTION_IND          = _cfdp.CfdpTransactionInd
    CFDP_EOF_SENT_IND             = _cfdp.CfdpEofSentInd
    CFDP_TRANSACTION_FINISHED_IND = _cfdp.CfdpTransactionFinishedInd
    CFDP_METADATA_RECV_IND        = _cfdp.CfdpMetadataRecvInd
    CFDP_FILE_SEGMENT_IND         = _cfdp.CfdpFileSegmentRecvInd
    CFDP_EOF_RECV_IND             = _cfdp.CfdpEofRecvInd
    CFDP_SUSPENDED_IND            = _cfdp.CfdpSuspendedInd
    CFDP_RESUMED_IND              = _cfdp.CfdpResumedInd
    CFDP_REPORT_IND               = _cfdp.CfdpReportInd
    CFDP_FAULT_IND                = _cfdp.CfdpFaultInd
    CFDP_ABANDONED_IND            = _cfdp.CfdpAbandonedInd
    CFDP_ALL_IND                  = 100

@unique
class CfdpConditionEnum(IntEnum):
    """ CFDP condition types. See ``help(CfdpConditionEnum)`` """
    CFDP_NO_ERROR                 = _cfdp.CfdpNoError
    CFDP_ACK_LIMIT_REACHED        = _cfdp.CfdpAckLimitReached
    CFDP_KEEP_ALIVE_LIMIT_REACHED = _cfdp.CfdpKeepaliveLimitReached
    CFDP_INVALID_TRANS_MODE       = _cfdp.CfdpInvalidTransmissionMode
    CFDP_FILESTORE_REJECT         = _cfdp.CfdpFilestoreRejection
    CFDP_CHECKSUM_FAIL            = _cfdp.CfdpChecksumFailure
    CFDP_FILESIZE_ERROR           = _cfdp.CfdpFileSizeError
    CFDP_NACK_LIMI_REACHED        = _cfdp.CfdpNakLimitReached
    CFDP_INACTIVITY_DETECTED      = _cfdp.CfdpInactivityDetected
    CFDP_INVALID_FILE_STRUCT      = _cfdp.CfdpInvalidFileStructure
    CFDP_CHECK_LIMIT_REACHED      = _cfdp.CfdpCheckLimitReached
    CFDP_SUSPED_REQUESTED         = _cfdp.CfdpSuspendRequested
    CFDP_CANCEL_REQUESTED         = _cfdp.CfdpCancelRequested

@unique
class CfdpFileStatusEnum(IntEnum):
    """ CFDP file status enumeration. See ``help(CfdpFileStatusEnum)`` """
    CFDP_FILE_DISCARDED  = _cfdp.CfdpFileDiscarded
    CFDP_FILE_REJECTED   = _cfdp.CfdpFileRejected
    CFDP_FILE_RETAINED   = _cfdp.CfdpFileRetained
    CFDP_FILE_UNREPORTED = _cfdp.CfdpFileStatusUnreported

@unique
class CfdpDeliverCodeEnum(IntEnum):
    """ CFDP delivery code enumeration. See ``help(CfdpDeliverCodeEnum)`` """
    CFDP_DATA_COMPLETE   = _cfdp.CfdpDataComplete
    CFDP_DATA_INCOMPLETE = _cfdp.CfdpDataIncomplete


import math
import os
import sys
from pathlib import Path
from setuptools import setup, find_packages, Extension

# --- PyION Build Logic ---
# Grabbing environment variables for C-extension build
ion_path = os.environ.get('ION_HOME')
bp_version = os.environ.get('PYION_BP_VERSION', 'BPv7') # Default to v7
lib_path = os.environ.get('LD_LIBRARY_PATH')

def get_extensions():
    actual_ion_path = ion_path
    
    # 1. Hybrid Auto-detection if ION_HOME is missing
    if not actual_ion_path:
        # Probe common system paths directly (bypasses PATH limitations in isolated environments)
        probes = ["/usr/local", "/usr"]
        for p in probes:
            if os.path.exists(os.path.join(p, "include", "ion.h")):
                actual_ion_path = p
                sys.stderr.write(f"INFO: Auto-detected ION in {p} via direct probe\n")
                break
        
        # Fallback to shutil.which if probes fail
        if not actual_ion_path:
            import shutil
            ionadmin_path = shutil.which("ionadmin")
            if ionadmin_path:
                if "/usr/local" in ionadmin_path:
                    actual_ion_path = "/usr/local"
                else:
                    actual_ion_path = str(Path(ionadmin_path).parent.parent)
                sys.stderr.write(f"INFO: Auto-detected ION: {actual_ion_path} via PATH\n")

    if not actual_ion_path:
        sys.stderr.write("WARNING: ION-DTN headers not found. Skipping C extensions.\n")
        sys.stderr.write("To force build, set ION_HOME=/path/to/ion or install via 'emion setup'.\n")
        return []

    # Paths for ION APIs
    ion_p = Path(actual_ion_path)
    is_system = actual_ion_path in ["/usr", "/usr/local"]
    
    if not lib_path:
        ion_inc = ion_p / 'include' if is_system else Path('/usr/local/include')
        ion_lib = ion_p / 'lib' if is_system else Path('/usr/local/lib')
    else:
        first_lp = lib_path.split(':')[0]
        lp = Path(first_lp)
        ion_inc = lp.parent / 'include'
        ion_lib = lp

    if is_system:
        ici_inc = ion_inc
        ici_lib = ion_lib
        ltp_lib = ion_lib
        cfdp_inc = ion_inc
        cfdp_lib = ion_lib
        bp_inc = ion_inc
        bp_lib = ion_lib
        bp_saga = ion_inc 
    else:
        ici_inc = ion_p / 'ici' / 'include'
        ici_lib = ion_p / 'ici' / 'library'
        ltp_lib = ion_p / 'ltp' / 'library'
        cfdp_inc = ion_p / 'cfdp' / 'include'
        cfdp_lib = ion_p / 'cfdp' / 'library'
        bp_inc = ion_p / (bp_version.lower()) / 'include'
        bp_lib = ion_p / (bp_version.lower()) / 'library'
        bp_saga = ion_p / (bp_version.lower()) / 'saga'

    # Machine type bit count
    hw_type = int(math.log2(sys.maxsize)) + 1
    ds = 3 if hw_type == 64 else (2 if hw_type == 31 else 1)

    compile_args = [
        '-g', '-Wall', '-O0', '-Wl,--no-undefined', '-Wno-undef',
        f'-DSPACE_ORDER={ds}', '-fPIC', '-Wno-unused-function',
        '-Wno-strict-prototypes', '-Wno-discarded-qualifiers', '-Wno-unused-variable'
    ]
    c_macros = [('PY_SSIZE_T_CLEAN', None), ('PYION_BP_VERSION', bp_version)]

    base_dir = os.path.dirname(os.path.abspath(__file__))
    local_inc = os.path.join(base_dir, 'emion', 'pyion')

    # Standardize on ion_inc and local_inc for most extensions
    ext_list = [
        Extension('emion.pyion._bp',
                  include_dirs=[str(ion_inc), str(bp_inc), local_inc, base_dir],
                  libraries=['bp', 'ici', 'ltp', 'cfdp'],
                  library_dirs=[str(ion_lib), str(bp_lib)],
                  sources=['emion/pyion/_bp.c', 'emion/pyion/base_bp.c'],
                  extra_compile_args=compile_args, define_macros=c_macros),
        Extension('emion.pyion._cfdp',
                  include_dirs=[str(ion_inc), str(cfdp_inc), local_inc, base_dir],
                  libraries=['cfdp', 'ici'],
                  library_dirs=[str(ion_lib), str(cfdp_lib)],
                  sources=['emion/pyion/_cfdp.c', 'emion/pyion/base_cfdp.c'],
                  extra_compile_args=compile_args, define_macros=c_macros),
        Extension('emion.pyion._ltp',
                  include_dirs=[str(ion_inc), str(ltp_lib), local_inc, base_dir],
                  libraries=['ltp', 'ici', 'bp', 'cfdp'],
                  library_dirs=[str(ion_lib), str(ltp_lib)],
                  sources=['emion/pyion/_ltp.c', 'emion/pyion/base_ltp.c'],
                  extra_compile_args=compile_args, define_macros=c_macros),
        Extension('emion.pyion._mem',
                  include_dirs=[str(ion_inc), str(ici_inc), local_inc, base_dir],
                  libraries=['ici', 'bp', 'cfdp', 'ltp'],
                  library_dirs=[str(ion_lib), str(ici_lib)],
                  sources=['emion/pyion/_mem.c', 'emion/pyion/base_mem.c'],
                  extra_compile_args=compile_args, define_macros=c_macros),
        Extension('emion.pyion._mgmt',
                  include_dirs=[str(ion_inc), str(bp_inc), str(bp_saga), str(ici_inc), local_inc, base_dir],
                  libraries=['ici', 'bp', 'ltp', 'cfdp'],
                  library_dirs=[str(ion_lib), str(bp_lib), str(ici_lib)],
                  sources=['emion/pyion/_mgmt.c'],
                  extra_compile_args=compile_args, define_macros=c_macros),
    ]
    return ext_list
    return ext_list

setup(
    name="emion",
    version="0.5.3",
    author="EmION Team",
    author_email="dev@info-gallary.com",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "emion": ["dashboard/static/*"],
    },
    ext_modules=get_extensions(),
    entry_points={
        "console_scripts": [
            "emion=emion.cli:main",
        ],
    },
    install_requires=[],
    extras_require={
        "dashboard": ["fastapi", "uvicorn", "websockets"],
    },
)
